/***************************************************************************
 *   Copyright (C) 2004 by Stephen McInerney                               *
 *   spm@stedee.id.au                                                      *
 *                                                                         *
 *   $Id: options.c 50 2005-08-11 09:46:21Z steve $
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * options.c
 *
 * Deal with Command line options
 ***************************************************************************
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * ModificationHistory:
 **********************
 * 03-Oct-2004 steve    Initial Creation
 * 09-Oct-2004 steve    Add "-d"
 *			Add version display
 * 16-Oct-2004 steve    Add "-t" timeout for visits - default is 30 minutes
 * 15-Apr-2005 steve    Minor Tidy Ups
 *                        Use ERRVPRINT Macro
 *                        use error.h exit codes
 * 12-Jun-2005 steve    Add query mode
 * 07-Jul-2005 steve    Fix use/setup of g_db_dirfilename & g_re_filter
 * 08-Jul-2005 steve    Add --cleanup option to remove older records
 *                      Add --unseen-since to query mode
 *                      Add --access-delta to query mode
 *
 ***************************************************************************
 ***************************************************************************/

#include "visitors.h"

/************************************************************************
 * 				GLOBALS					*
 ************************************************************************/
char *g_filename = NULL;                        /* Command Line specified file to open. */
time_t g_visittimeout = VISIT_TIME_SEP;         /* Default visitor timeout */

/* Global Flags - what do to'isims */
bool flag_raw_stats_only = false;               /* Output a TAB separated single line of raw, non-derived results only */
int db_cache_size = DB_CACHE_SIZE;


/************************************************************************
 * 				FUNCTIONS				*
 ************************************************************************/
void process_options (int, char *[]);           /* Process command line options */
void display_version (void);                    /* Display the Version Information */
void display_usage (int);                       /* Display the program Command Line Options, and exit */
void display_gnu_info (int);                    /* Display the GNU Disclaimer, and exit */


/************************************************************************
 * display_version	                                                *
 *                                                                      *
 * Display the Version Information					*
 ************************************************************************/
void
display_version (void)
{
    printf ("%s\n", PACKAGE_STRING);
    printf ("Copyright (C) %s Stephen McInerney\n\n", "2004, 2005");
}


/************************************************************************
 * display_usage	                                                *
 *                                                                      *
 * Display the Command Line Use and Exit with the passed exit_code	*
 ************************************************************************/
void
display_usage (int exit_code)
{
    display_version ();
    printf ("Usage: visitors [OPTION]...\n");
    printf ("  -c, --cache=SIZE           Size of the In Memory Cache to create in Mb\n");
    printf ("                               (default = %d)\n", DB_CACHE_SIZE);
    printf ("  -d, --database=FILE        Database File to store results in\n");
    printf ("                               (default = %s)\n", DATABASE);
    printf ("  -f, --file=FILE            Log File to process\n");
    printf ("                               if not set will use STDIN\n");
    printf ("  -F, --filter=RegExp        The Perl RE to use to filter what is NOT a Page\n");
    printf ("                               (default = \"%s\")\n", PATTERN_IGNORE_LIST);
    printf ("  -h, --help                 Display this message\n");
    printf ("  -s, --simple               Simple Results Output. Rest can be derived\n");
    printf ("  -t, --timeout=VALUE        Set timeout period for a visit, in seconds\n");
    printf ("                               (default = %d)\n", VISIT_TIME_SEP);
    printf ("  -v, --verbose              Level of verbosity, Multiple v's will increase\n");
    printf ("  -V, --version              Display the version information and exit\n");
    printf ("\n");
    printf ("  --%s=DAYS             Delete old entries that last visited over this\n", CMD_CLEANUP);
    printf ("                               many days ago. (NO Default)\n");
    printf ("\n");
    printf ("  --%s                    Go into Query mode\n", CMD_QUERY);
    printf ("  --%s=DAYS        Display the number of visitors per day who\n", CMD_Q_UNSEEN_SINCE);
    printf ("                               have not been seen in this many days.\n");
    printf ("  --%s             Display the number of visitors per number\n", CMD_Q_ACCESS_DELTA);
    printf ("                               of days between first and most recent visits.\n");


    exit (exit_code);
}


/************************************************************************
 * display_gnu_info	                                                *
 *                                                                      *
 * Display the GNU standard disclaimer					*
 ************************************************************************/
void
display_gnu_info (int exit_code)
{
    display_version ();
    printf ("%s\n%s%s%s\n%s\n%s\n",
            "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A",
            "PARTICULAR PURPOSE.  You may redistribute copies of ", PACKAGE_NAME, " under the terms", "of the GNU General Public License.", "For more information about these matters, see the file named COPYING.");
    exit (exit_code);
}


/************************************************************************
 * process_options                                                      *
 *                                                                      *
 * Process the individual commandline arguments                         *
 ************************************************************************/
void
process_options (int argc, char *argv[])
{
    int opt;
    int option_index = 0;

    const char *short_options = "c:d:f:F:hst:vV?";
    const struct option long_options[] = {
        {"cache", 1, NULL, 'c'},
        {"database", 1, NULL, 'd'},
        {"file", 1, NULL, 'f'},
        {"help", 0, NULL, 'h'},
        {"simple", 0, NULL, 's'},
        {"timeout", 1, NULL, 't'},
        {"verbose", 0, NULL, 'v'},
        {"version", 0, NULL, 'V'},
        {"filter", 1, NULL, 'F'},
        {CMD_CLEANUP, 1, NULL, 0},
        {CMD_QUERY, 0, NULL, 0},
        {CMD_Q_UNSEEN_SINCE, 1, NULL, 0},
        {CMD_Q_ACCESS_DELTA, 0, NULL, 0},
        {NULL, 0, NULL, 0}
    };

    do {
        opt = getopt_long (argc, argv, short_options, long_options, &option_index);
        switch (opt) {
        case 0:
            /*
            printf ("option %s", long_options[option_index].name);
            if (optarg)
                printf (" with arg %s", optarg);
            printf ("\n");
            */
            if (strncmp (CMD_CLEANUP, long_options[option_index].name, strlen (CMD_CLEANUP)) == 0) {
                sscanf (optarg, "%d", &g_cleanup_mode);
            } else if (strncmp (CMD_QUERY, long_options[option_index].name, strlen (CMD_QUERY)) == 0) {
                flag_query_mode = true;
            } else if (strncmp (CMD_Q_UNSEEN_SINCE, long_options[option_index].name, strlen (CMD_Q_UNSEEN_SINCE)) == 0) {
                sscanf (optarg, "%d", &g_qry_unseen_since);
            } else if (strncmp (CMD_Q_ACCESS_DELTA, long_options[option_index].name, strlen (CMD_Q_ACCESS_DELTA)) == 0) {
                flag_qry_access_delta = true;
            }
            break;
        case 'c':
            sscanf (optarg, "%d", &db_cache_size);
            break;
        case 'd':
            strncpy (g_db_dirfilename, optarg, strlen (optarg) + 1);
            break;
        case 'h':
            display_usage (V_EXIT_OK);
            break;
        case 'f':
            g_filename = optarg;
            break;
        case 'F':
            strncpy (g_re_filter, optarg, strlen (optarg) + 1);
            break;
        case 's':
            flag_raw_stats_only = true;
            break;
        case 't':
            sscanf (optarg, "%ld", &g_visittimeout);
            break;
        case 'v':
            g_verbosity++;
            if (g_verbosity > VERBOSE_MAX) {
                g_verbosity = VERBOSE_MAX;
                ERRVPRINT (VERBOSE0, msg_W_exceeded_verbosity, "\n");
            }
            break;
        case 'V':
            display_gnu_info (V_EXIT_OK);
            break;
        case '?':
            display_usage (V_EXIT_OK);
            break;
        case -1:
            break;
        default:
            exit (V_EXIT_INVALID_PARAMS);
        }
    } while (opt != -1);
    VPRINT (VERBOSE2, "Commandline Settings:%s\n", "");
    VPRINT (VERBOSE2, "  DB Cache Size: %d Mb\n", db_cache_size);
    VPRINT (VERBOSE2, "  DB File:       %s\n", g_db_dirfilename);
    VPRINT (VERBOSE2, "  Log File:      %s\n", g_filename);
    VPRINT (VERBOSE2, "  TimeOut:       %ld seconds\n", g_visittimeout);
    VPRINT (VERBOSE2, "  Filter:        %s\n", g_re_filter);
    VPRINT (VERBOSE2, "  Cleanup:       %d days\n", g_cleanup_mode);
}

/************************************************************************
 ************************************************************************
 *			END OF FILE					*
 ************************************************************************
 ************************************************************************/
