/***************************************************************************
 *   Copyright (C) 2004 by Stephen McInerney                               *
 *   spm@stedee.id.au                                                      *
 *                                                                         *
 *   $Id$
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * queries.c
 *
 * Do static analysis of the visitors DB
 ***************************************************************************
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * ModificationHistory:
 **********************
 * 12-Jun-2005 steve    Initial Creation
 *
 ***************************************************************************
 ***************************************************************************/

#include "visitors.h"

/************************************************************************
 *                              GLOBALS                                 *
 ************************************************************************/
bool flag_query_mode = false;                   /* Run in Query mode - additional flags for more details */
int g_qry_unseen_since = 0;                     /* Query: Shows numbers of visitors unseen in the last X days */
bool flag_qry_access_delta = false;             /* Query: number of visitors per number of days between first and most recent visits. */


/************************************************************************
 *                              FUNCTIONS                               *
 ************************************************************************/
void do_unseen_since (DBC *);
void do_access_delta (DBC * );


/************************************************************************
 * do_queries                                                           *
 *                                                                      *
 * Do Various Queries on the DB
 ************************************************************************
 *  db_ptr      :- double pointer to the DB structure                   *
 ************************************************************************/
void
do_queries (DB ** db_ptr                        /* The database to open */
    )
{

    /**********************************************************************/
    DBC *dbcurs_ptr = NULL;                     /* DB Pointer for cursor'ing */
    int db_rtn = 0;                             /* return value from DBD Get */

    /**********************************************************************/

    VPRINT (VERBOSE3, "%s\n", "Do Queries!");

    /* Acquire a cursor for the database. */
    if ((db_rtn = (*db_ptr)->cursor (*db_ptr, NULL, &dbcurs_ptr, 0)) != 0) {
        (*db_ptr)->err (*db_ptr, db_rtn, "DB->cursor");
        ERRVPRINT (VERBOSE0, msg_F_db_cursor_create, "");
        exit (V_EXIT_DB_CURSOR_CREATION);
    }
    VPRINT (VERBOSE5, "dbcurs_ptr == %p", dbcurs_ptr);

    if (g_qry_unseen_since > 0) {
        do_unseen_since (dbcurs_ptr);
    } else if (flag_qry_access_delta == true) {
        do_access_delta (dbcurs_ptr);
    }

    db_rtn = dbcurs_ptr->c_close (dbcurs_ptr);
}


/************************************************************************
 * do_unseen_since                                                      *
 *                                                                      *
 * Display the count of visitors per day not seen since in a given      *
 *  number of days.                                                     *
 *                                                                      *
 * Assumes that g_qry_unseen_since has a value greater than 0           *
 ************************************************************************
 *  dbcurs_ptr      :- pointer to an already opened cursoring DB        *
 ************************************************************************/
void
do_unseen_since (DBC * dbcurs_ptr)
{
    /**********************************************************************/
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    visit_data_type *visit_data = NULL;
    int db_rtn = 0;                             /* return value from DBD Get */

    time_t days_end;                            /* Holds the time at the end of the current day (23:59:59) */
    time_t day_index;                           /* Used for indexing back on a day by day basis */

    unsigned int i;
    char str_time[20];                          /* Temp holder for the ascii rep. of the day_index */
    struct tm *indexed_day_tm;                  /* matched to str_time - this holds the time value */

    unsigned long total = 0;                    /* Grand total of records that haven't been seen since */
    unsigned long *ary_ptr_unseen_list = NULL;  /* A array to be created
                                                   Holds the number of unseen since visitors per day
                                                   Each "cell" == a given day, where cell 0 = today
                                                   and hence cell 0 is not used.
                                                   Cell 1 = yesterday and so on. */

    /**********************************************************************/

    VPRINT (VERBOSE3, "%s\n", "Do Unseen Since!");

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    days_end = time_end_today();
    day_index = days_end;

    ary_ptr_unseen_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);

    while ((db_rtn = dbcurs_ptr->c_get (dbcurs_ptr, &dbt_key, &dbt_data, DB_NEXT)) == 0) {
        visit_data = (visit_data_type *) dbt_data.data;
        if (difftime (days_end, visit_data->last_access) / DAY_IN_SECONDS >= g_qry_unseen_since) {
            if (ary_ptr_unseen_list + ((days_end - visit_data->last_access) / DAY_IN_SECONDS) == 0) {
                printf ("Day: %ld\n", (days_end - visit_data->last_access) / DAY_IN_SECONDS);
            }
            (*(ary_ptr_unseen_list + ((days_end - visit_data->last_access) / DAY_IN_SECONDS)))++;
        }
    }

    printf ("Date         Nbr Unseen Visitors\n");
    for (i = 0; i < YEARS_TO_TRACK * DAYS_IN_YEAR; i++) {
        day_index -= DAY_IN_SECONDS;
        if (*(ary_ptr_unseen_list + i) > 0) {
            indexed_day_tm = localtime (&day_index);
            strftime (str_time, 20, "%d-%b-%Y", indexed_day_tm);

            printf ("%s      %15lu\n", str_time, *(ary_ptr_unseen_list + i));
            total += *(ary_ptr_unseen_list + i);
        }
    }
    printf ("================================\n");
    printf ("Total:           %15lu\n", total);

    XFREE (ary_ptr_unseen_list);
}


/************************************************************************
 * do_access_delta                                                      *
 *                                                                      *
 * Display various numbers about the delta in days between a first and  *
 *  most recent visit.                                                  *
 *                                                                      *
 ************************************************************************
 *  dbcurs_ptr      :- pointer to an already opened cursoring DB        *
 ************************************************************************/
void
do_access_delta (DBC * dbcurs_ptr)
{
    /**********************************************************************/
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    visit_data_type *visit_data = NULL;
    int db_rtn = 0;                             /* return value from DBD Get */

    time_t days_end;                            /* Holds the time at the end of the current day (23:59:59) */
    unsigned int day_delta;

    unsigned int i;
    unsigned int max_delta = 0;

    unsigned long *ary_ptr_delta_list = NULL;   /* A array to be created
                                                   Holds the number of visitors per day with a given delta
                                                   Cell 0 = no delta
                                                   Cell 1 = 1 days delta and so on */
    unsigned long *ary_ptr_delta_visits_list = NULL;   /* ditto - number of visits */
    unsigned long *ary_ptr_delta_pages_list = NULL;   /* ditto - number of pages */
#if 0
    unsigned long *ary_ptr_delta_cookie_list = NULL;   /* ditto - but verified via cookie, visitors per day */
    unsigned long *ary_ptr_delta_cookie_visits_list = NULL;   /* verified visits */
    unsigned long *ary_ptr_delta_cookie_pages_list = NULL;   /* verified pages */
#endif
    /**********************************************************************/

    VPRINT (VERBOSE2, "%s\n", "Do Access Delta!");

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    days_end = time_end_today();

    ary_ptr_delta_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);
    ary_ptr_delta_pages_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);
    ary_ptr_delta_visits_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);

    /*
    ary_ptr_delta_cookie_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);
    ary_ptr_delta_cookie_visits_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);
    ary_ptr_delta_cookie_pages_list = XMALLOC (unsigned long, YEARS_TO_TRACK * DAYS_IN_YEAR);
    */

    while ((db_rtn = dbcurs_ptr->c_get (dbcurs_ptr, &dbt_key, &dbt_data, DB_NEXT)) == 0) {
        visit_data = (visit_data_type *) dbt_data.data;
        day_delta = difftime (visit_data->last_access, visit_data->first_access) / DAY_IN_SECONDS;
        (*(ary_ptr_delta_list + day_delta))++;
        (*(ary_ptr_delta_visits_list + day_delta)) += visit_data->nbr_visits;
        (*(ary_ptr_delta_pages_list + day_delta)) += visit_data->nbr_pages;
        
        /* Increase definite delta - known due to same cookie */
        /*
        if (visit_data->good_cookie == true) {
            (*(ary_ptr_delta_cookie_list + day_delta))++;
            (*(ary_ptr_delta_cookie_visits_list + day_delta)) += visit_data->nbr_visits;
            (*(ary_ptr_delta_cookie_pages_list + day_delta)) += visit_data->nbr_pages;
        }
        */

        if (day_delta > max_delta) {
            max_delta = day_delta;
        }
    }

    /*
    printf (" Days         Nbr      Avg     Avg         Nbr            Avg     Avg\n");
    printf ("Delta    Visitors   Visits   Pages   Definites         Visits   Pages\n");
    */
    printf (" Days         Nbr      Avg     Avg\n");
    printf ("Delta    Visitors   Visits   Pages\n");
    for (i = 0; i <= max_delta; i++) {
        printf (" %4u     %7lu   ", i, *(ary_ptr_delta_list + i));
        if (*(ary_ptr_delta_list + i) > 0) {
            printf ("%6.1f  %6.1f", (float) *(ary_ptr_delta_visits_list + i) / (float) *(ary_ptr_delta_list + i), (float) *(ary_ptr_delta_pages_list + i) / (float) *(ary_ptr_delta_list + i));
            /* printf ("     %7lu  %3.1f%%  ", *(ary_ptr_delta_cookie_list + i), ((float) *(ary_ptr_delta_cookie_list + i) / (float) *(ary_ptr_delta_list + i)) * 100.0); */
        } else {
            printf ("%6.1f  %6.1f", 0.0, 0.0);
            /* printf ("     %7lu  %3.1f%%  ", *(ary_ptr_delta_cookie_list + i), 0); */
        }
        printf ("\n");

        /*
        if (*(ary_ptr_delta_cookie_list + i) > 0) {
            printf ("%6.1f  %6.1f\n", (float) *(ary_ptr_delta_cookie_visits_list + i) / (float) *(ary_ptr_delta_cookie_list + i),  (float) *(ary_ptr_delta_cookie_pages_list + i) / (float) *(ary_ptr_delta_cookie_list + i));
        } else {
            printf ("%6.1f  %6.1f\n", 0.0, 0.0);
        }
        */
    }

    XFREE (ary_ptr_delta_list);
    XFREE (ary_ptr_delta_visits_list);
    XFREE (ary_ptr_delta_pages_list);

    /*
    XFREE (ary_ptr_delta_cookie_list);
    XFREE (ary_ptr_delta_cookie_visits_list);
    XFREE (ary_ptr_delta_cookie_pages_list);
    */
}


/************************************************************************
 ************************************************************************
 *                      END OF FILE                                     *
 ************************************************************************
 ************************************************************************/
