/***************************************************************************
 *   Copyright (C) 2004 by Stephen McInerney                               *
 *   spm@stedee.id.au                                                      *
 *                                                                         *
 *   $Id: visitors.c 47 2005-07-12 11:42:15Z steve $
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * visitors.c
 *
 * Track Visitors in Web Logs
 * Assumes that mod_usertrack is enabled, or some other Cookie is logged
 *   better accuracy.
 * Will also count via IP/Browser Combo as well - if cookies blocked.
 *
 *
 * NOTE:
 * An "Old Visitor" is a visitor that has been seen BEFORE this run
 *   This assumes that this is at least the 2nd run of this program against new data
 *   And that the cookie DB has been retained between runs (not deleted)
 *   On the first ever run, Old Visitors and related stats should = ZERO (0)
 *
 * A "New Visitor" is the converse, a visitor for whom this is the first sighting - ever.
 *
 *
 * Formatting Style:
 * *****************
 * indent -kr -bad -bap -bbo -nbc -br -brs -c49 -cd49 -i4 -ce -bli4 -cp17 -l250 -nut -pcs -psl
 ***************************************************************************
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * Logic:
 *
 * A new visit is determined by the same cookie|Address/browser, but last seen over g_visittimeout seconds ago (30 minutes by default)
 * Primarily look at the cookie field. Decision matrix follows.
 * Visitors are ONLY counted by page hits, NOT any hit.
 *   May want to count "visitors" who don't actually ask for a page, image etc only???
 * Do we count 4xx's or 5xx's as new visitors???
 *   Assume yes, MUST provide alternate counts.
 *   One that includes, One that doesn't
 *
 * If we have ::--
 * (a) Cookies are the same: Ignore Address/Browser Combo
 *      (i)     Same visitor
 *      The odd's against two visitors having the same cookie are quite long, accept risk
 * (b) Cookies are different; Address/Browser is the same:
 *      (i)     Multiple visitors, accepting cookies - over time will resolve down to case (a)
 *      (ii)    Multiple visitors, dropping cookies
 *      (iii)   Same visitor, dropping Cookies
 *      (iv)    Same visitor, only looking at pages - eg robot style, MSD, UNSW
 * (c) Cookies are different; Address/Browser differs:
 *      (i)     Multiple visitors, accepting cookies - over time will resolve down to case (a)
 *      (ii)    Multiple visitors, dropping cookies
 *      (iii)   Same visitor, blocking cookies, spread proxy (eg AOL user, blocking cookies)
 *      (iv)    eg GoogleBot. - Same visitor, effectively, but from multiple sources
 *
 * if #hits == 1 - is same visitor, blocking cookies.
 * --+ (a) is easy - just look at cookie, ignore Address/Browser - delete any addr/browser data for this machine
 * --+ (b) Look at hits.
 *              1.      If multiple (not just pages) hits for same cookie, then becomes case (a)
 *                       ie. We actually do have multiple same cookies
 *              2.      If 1 hit per cookie, we have no way of separating out ii, iii or iv.
 *                       Assume same visitor. ie "normal" behaviour - Addr/Browser only
 * --+ (c) Initially assume differing visitors.
 *              1.      We could look at known RE's in the Browser field, and deduce a single visitor
 *                       where we would otherwise have > 1 visitors.
 *
 ***************************************************************************
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * TODO:
 *
 * *    Number Pages per Visit/Visitor
 * *    Break down visits/Visitors per day/hour whatever
 * *    Visit Lengths
 * *    Track Visitors by Country/Domain
 *
 *
 * BUGS:
 * *    test case 11 - two (or more) good cookies, arriving close together
 *       ie no intervening "image" hits to verify the cookies
 *       This will fail to register the 1st visitor as a separate entity correctly
 *
 ***************************************************************************
 ***************************************************************************/

/***************************************************************************
 ***************************************************************************
 * Modification History:
 **********************
 * 19-Jun-2004 steve    Initial Creation
 * 07-Nov-2004 tony     DB Pointer changes - get rid of Global
 * 19-Nov-2004 steve    Tidy up tony's patch
 * 14-Mar-2005 steve    Correct Cookie Cleanup - missing time check
 * 15-Apr-2005 steve    More cleanups.
 *                        Remove remaining DB ptr globals
 *                        Fix drop_long_line only dropping from stdin
 *                          arguments for buffer and FILE
 *                        use XMALLOC macro
 *                        use ERRVPRINT macro instead fprintf (stderr...
 *                        Tidy up end results display
 *                        Add total log lines counter - assists log file error finding
 *                        Remove unnecessary malloc of DB Ptr in main
 *                        Re-correct Cookie cleanup - time check was broken. Badly.
 * 07-Jul-2005 steve    Fix set/use of g_db_dirfilename & g_re_filter
 *                      Fix function declarations for display_results
 * 08-Jul-2005 steve    Add cleanup old records functionality
 *                      Add function time_end_today
 * 09-Jul-2005 steve    Shuffle VERBOSE levels
 *
 ***************************************************************************
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif          /* HAVE_CONFIG_H */

#include "visitors.h"

/************************************************************************
 *                              GLOBALS                                 *
 ************************************************************************/
cookie_type current_cookie;                     /* Current cookie */
address_browser_type current_address_browser;   /* Current Address/Browser */
time_t run_start_time = 0;                      /* Holds the first log entry time - this is what determines new vs repeat visitors */

//int rtn_db;                                   /* Value from DB calls; Used for program exit if necessary */

/* Generic Counters */
unsigned long int total_lines = 0;              /* How many log lines in total */
unsigned long int bad_lines = 0;                /* How many dud log lines did we get? */

/* The Various Results */
unsigned long int nbr_visitors = 0;             /* How many visitors/people have we seen? */
unsigned long int nbr_visits = 0;               /* How many visits have we seen? */
unsigned long int nbr_new_visitors = 0;         /* How many new visitors/people have we seen? */
unsigned long int nbr_pages_viewed = 0;         /* How many pages have all our visitors accessed? */
unsigned long int nbr_pages_new = 0;            /* How many pages have all our NEW visitors accessed? */
unsigned long int nbr_pages_old = 0;            /* How many pages have all our OLD visitors accessed? */
unsigned long int nbr_visits_new = 0;           /* How many visits have all our NEW visitors done? */
unsigned long int nbr_visits_old = 0;           /* How many visits have all our OLD visitors done? */
unsigned long int successful_cookies = 0;       /* How many successful cookies? */
unsigned long int cleaned_cookies = 0;          /* How many cookies did we clean/remove? */

time_t current_time = 0;                        /* The current log time */

char g_db_dirfilename[MAX_FILENAME_LENGTH] = DATABASE;  /* File name for the Database */
char g_re_filter[MAX_RE_LENGTH] = PATTERN_IGNORE_LIST;  /* Hold the URL filter RE */

cookie_list_element *new_cookies_list = NULL;   /* Linked list of new cookies from this run */

int g_cleanup_mode = 0;                         /* Number of days beyond which older records will be deleted */


/************************************************************************
 *                              FUNCTIONS                               *
 ************************************************************************/
void drop_long_lines (FILE *, char *);          /* Strip Log Lines that are too big */

inline int cookie_exists (DB **, cookie_type *);        /* Does this cookie exist already ?? */
inline int address_browser_exists (DB **, address_browser_type *);

inline void store_cookie (DB **, cookie_type *);        /* Store a cookie record */
inline void store_address_browser (DB **, address_browser_type *);      /* Store an address/browser record */

unsigned int cleanup_excess_cookies (DB **);    /* Delete unused cookies from the DB */

void display_results (void);                    /* Display the results */
void cleanup_old_records (DB **);               /* Remove old visitors */

/************************************************************************
 ************************************************************************
 ***                            MAIN                                  ***
 ************************************************************************
 ************************************************************************/
/* MAIN */
int
main (int argc, char *argv[])
{
    /**********************************************************************/
    char re_main[MAX_RE_LENGTH] = PATTERN_NCSA_COOKIE;  /* Hold the primary log splitter RE */
    const char *error;                          /* RE error pointer, offset */
    int erroffset;                              /* RE error value */
    int ovector[OVECCOUNT];                     /* RE substring offsets array */
    int rc;                                     /* RE Check return value */
    int copy_substr_rtn;                        /* RE Check return from pcre_copy_substring */

    /* Log File handlers */
    FILE *input_file = NULL;                    /* Input File descriptor */
    char *fgets_rtn = NULL;                     /* Return value from fgets - check for NULL to exit main loop */


    char buffer[BUFSIZE];                       /* log buffer */
    int buffer_length;

    DB *visitor_db_ptr;                         /* Database Pointer */

    /* Various Strings - typically log file extracts */
    char str_time[SIZE_DATE_TIME];              /* String to hold the current time out of the current log line */
    struct tm time_rec;                         /* Gotta convert that string'ed time into a timerec first */
    char str_url[BUFSIZE];                      /* Substring holder for URL's */
    char str_address[SIZE_ADDRESS];             /* IP Addresses */
    char str_browser[SIZE_BROWSER];             /* Browser */

    int filtered_url;                           /* Is this a filtered URL? */

    /**********************************************************************/

    /*****************************
     *          BEGIN
     *****************************/
    process_options (argc, argv);

    VPRINT (VERBOSE2, "Pattern Log: %s\n", re_main);
    VPRINT (VERBOSE2, "Pattern Filter: %s\n", g_re_filter);
    VPRINT (VERBOSE3, "Compiling Regular Expressions%s\n", "");

    re_pattern_main = pcre_compile (re_main, 0, &error, &erroffset, NULL);
    if (re_pattern_main == NULL)
        re_compile_failed (erroffset, error, re_main);

    re_pattern_filter = pcre_compile (g_re_filter, 0, &error, &erroffset, NULL);
    if (re_pattern_filter == NULL)
        re_compile_failed (erroffset, error, g_re_filter);

    /* Is this a specified file? Open if so. */
    if (g_filename != NULL) {
        VPRINT (VERBOSE2, "Using File: %s\n", g_filename);
        input_file = fopen (g_filename, "r");
        if (input_file == NULL) {
            ERRVPRINT (VERBOSE0, "%s%s\n", msg_F_file_open, g_filename);
            exit (V_EXIT_BAD_FILE_OPEN);
        }
    }

    open_visitor_db (&visitor_db_ptr, g_db_dirfilename, db_cache_size);

    /* Do any non standard processing: Cleanups, Queries etc and EXIT! */
    if (g_cleanup_mode > 0) {
        cleanup_old_records (&visitor_db_ptr);
        close_visitor_db (&visitor_db_ptr);
        return (V_EXIT_OK);                     /* FIXME */
    } else if (flag_query_mode == true) {
        do_queries (&visitor_db_ptr);
        close_visitor_db (&visitor_db_ptr);
        return (V_EXIT_OK);                     /* FIXME */
    }


    VPRINT (VERBOSE3, "Beginning Main Loop%s\n", "");
    while (1) {
        if (input_file != NULL) {
            fgets_rtn = fgets (buffer, BUFSIZE, input_file);
        } else {
            fgets_rtn = fgets (buffer, BUFSIZE, stdin);
        }
        if (fgets_rtn == NULL) {
            /* Exit Main Loop at end of input */
            break;
        }

        VPRINT (VERBOSE5, "\t\tNext Line%s\n", "");

        total_lines++;

        buffer_length = (int) strlen (buffer);
        /* check for log line too long */
        if (strpbrk ("\n", buffer) == NULL) {
            drop_long_lines (input_file, buffer);
            bad_lines++;
            continue;
        }

        memset (&current_cookie, 0, sizeof (current_cookie));   /* Zero Current Cookie */

        rc = pcre_exec (re_pattern_main, NULL, buffer, buffer_length, 0, 0, ovector, OVECCOUNT);
        /* check for RE matching errors */
        if (rc <= 0) {
            re_check_errors (rc, total_lines, buffer);
            bad_lines++;
            continue;
        }
#if 0
        /* What to do about failures? Is this a new visit or not??? */
        status_start = buffer + ovector[2 * LF_NCSA_STATUS];
        if (*status_start == '4' || *status_start == '5') {
            /* Have a 4xx or 5xx error, next line */
            continue;
        }
#endif

        /* Extract and set the start time for all data presented on this run.
         * This "run_start_time" is thus used to determine new vs old visitors
         *  and achieve an accurate visitor count. */
        if (run_start_time == 0) {
            copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_NCSA_DATE_TIME, str_time, SIZE_DATE_TIME);
            if (copy_substr_rtn < 0) {
                error_substring_extract (re_main, buffer, LF_NCSA_DATE_TIME, copy_substr_rtn);
                ERRVPRINT (VERBOSE0, "%s\n", msg_F_run_start_datetime);
                exit (V_EXIT_PCRE_NO_SUBSTR);
            }
            strptime (str_time, DATE_TIME_FORMAT, &time_rec);
            run_start_time = mktime (&time_rec);
            VPRINT (VERBOSE3, "VISITORS START TIME: %s -> %ld\n", str_time, run_start_time);
        }

        /* Extract the URL substring */
        copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_NCSA_URL, str_url, BUFSIZE);
        if (copy_substr_rtn < 0) {
            error_substring_extract (re_main, buffer, LF_NCSA_URL, copy_substr_rtn);
            bad_lines++;
            continue;
        }
        filtered_url = is_filtered (str_url);

        /* Extract the cookie field */
        copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_CUSTM_COOKIE, current_cookie.cookie_key, SIZE_COOKIE);
        if (copy_substr_rtn < 0) {
            error_substring_extract (re_main, buffer, LF_CUSTM_COOKIE, copy_substr_rtn);
            bad_lines++;
            continue;
        }


        /* If this is a filtered URL - eg Image etc then we are not interested in doing date/time calc's
         * Instead, we use filtered URL's to assert that cookies are in fact ok or not */
        if (filtered_url == 0) {
            nbr_pages_viewed++;                 /* Increase the total page count */
            /* Do the time conversions from a text field to data */
            copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_NCSA_DATE_TIME, str_time, SIZE_DATE_TIME);
            if (copy_substr_rtn < 0) {
                ERRVPRINT (VERBOSE0, "%s\n", msg_F_vital_substring);
                exit (V_EXIT_PCRE_NO_SUBSTR);
            }
            memset (&time_rec, 0, sizeof (time_rec));   /* reset time_rec - failing to do so causes time comparison problems. Yuk */
            strptime (str_time, DATE_TIME_FORMAT, &time_rec);
            current_time = mktime (&time_rec);
            VPRINT (VERBOSE3, "**Date/Time: %s -> %ld\n", str_time, current_time);
            if (difftime (run_start_time, current_time) > 0) {
                ERRVPRINT (VERBOSE0, "%s Line: %ld  %ld > %ld = %.0f\n", msg_E_log_misordered, total_lines, current_time, run_start_time, difftime (run_start_time, current_time));
                /* ERRVPRINT (VERBOSE2, "\t%s\n", str_time); */
                bad_lines++;
                continue;
            }
            /* TZ issues??? */
        }


        /************************************************************************
         * Start doing the Logic Cases to discover a new visit/visitor          *
         ************************************************************************/
        if (cookie_exists (&visitor_db_ptr, &current_cookie) == 1) {
            VPRINT (VERBOSE3, "C: COOKIE DOES exist: %s\n", current_cookie.cookie_key);

            /********************************************************************
             * Ok, we have seen this cookie before, therefore we can continue to
             * rely on this cookie as being "ok", and hence base decisions
             * against this cookie only.
             *  - We can delete the corresponding address/browser entry.
             *  - This may be a new visit.
             *  - This may be the first visit of this visitor in this log run.
             *
             *
             * Logic Case: (a) (b) & (c)
             *
             * We have not yet filtered, so this could be a to-be "filtered" hit
             * This is ok, as this could assert a valid cookie!
             *
             * If Not a Filtered item:
             *   Old Visitor:
             *       ? possible new visit
             *       ? possible new visitor for this run
             *       eg Two page hits by the same person (page spider?)
             *     Check for new visit
             *       If new visit, check if new visitor for this run
             *         May have already counted this visitor in this run;
             *           check and decrement as necessary
             *     Update last access
             *     Update cookie record
             *
             * If a Filtered or Not Filtered item, but haven't set good_cookie:
             *   Same visitor, this cookie is good!
             *     Delete matching address/browser (Must have same cookie!)
             *     Set "good_cookie"
             *     Update cookie record
             ********************************************************************/

            /*   Prior visitor, new visit       */
            if (filtered_url == 0) {
                /* Ignore additional non page hits */
                VPRINT (VERBOSE3, "  C: Not Filtered.%s\n", "");
                if (difftime (current_time, current_cookie.data.last_access) > g_visittimeout) {
                    /* New Visit */
                    VPRINT (VERBOSE3, "    C: NEW VISIT: last time: %ld is > visit timeout: %ld\n", current_cookie.data.last_access + g_visittimeout, current_time);
                    /* So, is this a new VISITOR? */
                    if (difftime (run_start_time, current_cookie.data.last_access) > 0) {
                        /* Yup! */
                        VPRINT (VERBOSE3, "      C: NEW VISITOR: previous time: %ld is < run start time: %ld\n", current_cookie.data.last_access, run_start_time);
                        nbr_visitors++;
                        if (current_cookie.data.good_cookie == false) {
                            VPRINT (VERBOSE4, "      C: OLD VISITOR: Has already been counted. Decrementing visitors count.%s\n", "");
                            nbr_visitors--;
                        }
                    }
                    current_cookie.data.nbr_visits++;   /* Nbr visits increases for this cookie */
                    current_cookie.data.visit_start = current_time;     /* Reset visit start time to new visit */
                    nbr_visits++;

                    VPRINT (VERBOSE4, "    C: New or Old Visit?: is %ld >= %ld?\n", current_cookie.data.first_access, run_start_time);
                    if (difftime (current_cookie.data.first_access, run_start_time) >= 0) {
                        VPRINT (VERBOSE4, "      C: NEW VISIT INC: New Visitor: Increase Number New Visits.%s\n", "");
                        nbr_visits_new++;
                    } else {
                        VPRINT (VERBOSE4, "      C: OLD VISIT INC: Old Visitor: Increase Number Old Visits.%s\n", "");
                        nbr_visits_old++;
                    }
                }

                /* Is this a new visitor for this run? increase old or new page count accordingly */
                VPRINT (VERBOSE4, "  C: New or Old Page?: is %ld >= %ld?\n", current_cookie.data.first_access, run_start_time);
                if (difftime (current_cookie.data.first_access, run_start_time) >= 0) {
                    VPRINT (VERBOSE4, "    C: NEW PAGE INC: New Visitor: Increase Number New Pages.%s\n", "");
                    nbr_pages_new++;
                } else {
                    VPRINT (VERBOSE4, "    C: OLD PAGE INC: Old Visitor: Increase Number Old Pages.%s\n", "");
                    nbr_pages_old++;
                }

                current_cookie.data.nbr_pages++;
                current_cookie.data.last_access = current_time; /* Set last access time on pages ONLY */
                store_cookie (&visitor_db_ptr, &current_cookie);
            }

            if (current_cookie.data.good_cookie == false) {
                /* We've now seen this cookie twice; So this is a Known Good Cookie.
                 * Can be set via either a page hit or regular hit. */
                VPRINT (VERBOSE4, "  C: Set Good Cookie.%s\n", "");
                del_hash (&visitor_db_ptr, current_cookie.address_browser);     /* Can remove the initial address/browser combo */
                current_cookie.address_browser[0] = '\0';       /* "clear" the browser/address field. No longer needed. */
                current_cookie.data.good_cookie = true; /* This is a "good cookie" - multiple seeing's */
                successful_cookies++;
                /* Update this cookie */
                store_cookie (&visitor_db_ptr, &current_cookie);
            }

        } else {                                /* Cookie Does NOT exist. */
            VPRINT (VERBOSE4, "AB: COOKIE does NOT exist (yet...): %s\n", current_cookie.cookie_key);

            /* Logic Case: (b) & (c)            *
             *   or early (a)                   */

            /*
             * If a Filtered item:
             *  Have we seen this cookie?
             *    yes: == (a)
             *      We Should not be here - picked up above!
             *    no:
             *      Ignore, lack of info to decide one way or t'other.
             *
             * If Not a Filtered item:
             *  Create new cookie entry.  Set up for (a)
             *
             * Have we seen this Addr/Browser before?
             *   yes:
             *     Assume a std. Addr/Browser visitor
             *     Check for new visit timeouts
             *   no:
             *     Create new Addr/Browser entry.  Set up for (b) & (c)
             */

            /* Is this a page or an image? */
            if (filtered_url == 0) {
                /* Is a Page.
                 * Could be new cookie, or old address/browser (blocked cookie?)
                 * Assume new cookie in first instance */

                /* Need to extract and combine the Address/Browser Fields - ignored for cookie matches */
                memset (&current_address_browser, 0, sizeof (current_address_browser)); /* Zero Current Address/Browser */
                /* Address Field */
                copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_NCSA_ADDRESS, str_address, SIZE_ADDRESS);
                if (copy_substr_rtn < 0) {
                    error_substring_extract (re_main, buffer, LF_NCSA_ADDRESS, copy_substr_rtn);
                    bad_lines++;
                    continue;
                }
                /* Browser Type Field */
                copy_substr_rtn = pcre_copy_substring (buffer, ovector, rc, LF_NCSA_BROWSER, str_browser, SIZE_BROWSER);
                if (copy_substr_rtn < 0) {
                    error_substring_extract (re_main, buffer, LF_NCSA_BROWSER, copy_substr_rtn);
                    bad_lines++;
                    continue;
                }
                /* Combine the Address/Browser Combo */
                snprintf (current_address_browser.address_browser_key, SIZE_ADDRESS + SIZE_BROWSER, "%s%s", str_address, str_browser);

                VPRINT (VERBOSE3, "  AB: Not Filtered.%s\n", "");
                VPRINT (VERBOSE4, "  AB: Address: %s\tBrowser: %s\n", str_address, str_browser);

                /* Have we seen this address/browser before? */
                if (address_browser_exists (&visitor_db_ptr, &current_address_browser) == 1) {
                    /* Yes - have seen before. */
                    VPRINT (VERBOSE3, "    AB: IS OLD Browser: %s\n", current_address_browser.address_browser_key);
                    current_address_browser.data.nbr_pages++;
                    if (difftime (current_time, current_address_browser.data.last_access) > g_visittimeout) {
                        /* New Visit */

                        /* So, is this a new VISITOR? */
                        if (difftime (run_start_time, current_address_browser.data.last_access) > 0) {
                            /* Yup! */
                            VPRINT (VERBOSE3, "      AB: NEW VISITOR: previous time: %ld is < %ld\n", current_address_browser.data.last_access, run_start_time);
                            nbr_visitors++;
                        }
                        nbr_visits++;
                        current_address_browser.data.nbr_visits++;
                        current_address_browser.data.visit_start = current_time;
                        /* Is this a new visitor for this run? increase old or new visit count accordingly */
                        VPRINT (VERBOSE4, "    AB: New or Old Visit?: is %ld >= %ld?\n", current_address_browser.data.first_access, run_start_time);
                        if (difftime (current_address_browser.data.first_access, run_start_time) >= 0) {
                            VPRINT (VERBOSE4, "      AB: NEW VISIT INC: New Visitor: Increase Number New Visits.%s\n", "");
                            nbr_visits_new++;
                        } else {
                            VPRINT (VERBOSE4, "      AB: OLD VISIT INC: Old Visitor: Increase Number Old Visits.%s\n", "");
                            nbr_visits_old++;
                        }
                    }
                    /* Is this a new visitor for this run? increase old or new page count accordingly */
                    VPRINT (VERBOSE4, "    AB: New or Old Page?: is %ld >= %ld?\n", current_address_browser.data.first_access, run_start_time);
                    if (difftime (current_address_browser.data.first_access, run_start_time) >= 0) {
                        VPRINT (VERBOSE4, "      AB: NEW PAGE INC: New Visitor: Increase Number New Pages.%s\n", "");
                        nbr_pages_new++;
                    } else {
                        VPRINT (VERBOSE4, "      AB: OLD PAGE INC: Old Visitor: Increase Number Old Pages.%s\n", "");
                        nbr_pages_old++;
                    }
                } else {
                    VPRINT (VERBOSE3, "    AB: BRAND NEW VISITOR%s\n", "");
                    /* No. This is a new address/browser.
                     * We should pass thru here for EVERY new
                     * visitor from a new address/browser */

                    /* Increase visit/visitor counts */
                    nbr_visits++;
                    nbr_visitors++;
                    nbr_new_visitors++;
                    nbr_pages_new++;
                    nbr_visits_new++;

                    /* Set up base counts for the address/browser key */
                    current_address_browser.data.nbr_pages = 1;
                    current_address_browser.data.nbr_visits = 1;
                    current_address_browser.data.visit_start = current_time;
                    current_address_browser.data.first_access = current_time;
                    current_address_browser.data.is_cookie = false;

                    strncpy (current_cookie.address_browser, current_address_browser.address_browser_key, SIZE_ADDRESS + SIZE_BROWSER);

                    /* New cookie: set base counts */
                    current_cookie.data.nbr_pages = 1;
                    current_cookie.data.nbr_visits = 1;
                    current_cookie.data.visit_start = current_time;
                    current_cookie.data.last_access = current_time;
                    current_cookie.data.first_access = current_time;
                    current_cookie.data.is_cookie = true;
                    /* Add this new cookie */
                    store_cookie (&visitor_db_ptr, &current_cookie);
                }
                /* Either way, update the last_access time and total page views! */
                current_address_browser.data.last_access = current_time;
                store_address_browser (&visitor_db_ptr, &current_address_browser);
            }                                   /* end; filtered AB combo */
        }                                       /* if/else cookie exists */
    }                                           /* while read file/stdin */

    cleaned_cookies = cleanup_excess_cookies (&visitor_db_ptr); /* Delete all "unused" cookie entries */

    display_results ();

    close_visitor_db (&visitor_db_ptr);
    /* free (visitor_db_ptr); */
    return (V_EXIT_OK);                         /* FIXME */
}


/************************************************************************
 ************************************************************************
 *                              END                                     *
 ************************************************************************
 ************************************************************************/


/************************************************************************
 * drop_long_lines                                                      *
 *                                                                      *
 * Strip Log Lines that are too big                                     *
 *                                                                      *
 * Arguments:                                                           *
 * FILE * file_input            The input File. If NULL, use STDIN.     *
 * char *buffer_ptr             Pointer to the buffer to use.           *
 *                              This buffer is max size: BUFSIZE        *
 ************************************************************************/
void
drop_long_lines (FILE * file_input, char *buffer_ptr)
{
    /**********************************************************************/
    char *fgets_rtn = NULL;

    /**********************************************************************/

    if (buffer_ptr == NULL) {
        ERRVPRINT (VERBOSE0, "%s\n  Exiting at Log Line: %lu\n", msg_F_buffer_missing, total_lines);
        exit (V_EXIT_MISSING_BUFFER);
    }

    buffer_ptr[BUFCUTOFF] = '\0';
    ERRVPRINT (VERBOSE1, "Drop Long Line: #%lu  %s:\n\t%s...\n", total_lines, msg_W_line_too_big, buffer_ptr);

    while (1) {
        if (file_input == NULL) {
            fgets_rtn = fgets (buffer_ptr, BUFSIZE, stdin);
        } else {
            fgets_rtn = fgets (buffer_ptr, BUFSIZE, file_input);
        }
        if (strpbrk ("\n", buffer_ptr) != NULL) {
            break;
        }
    }
}


/************************************************************************
 * display_results                                                      *
 *                                                                      *
 * Display the results of the run                                       *
 * Requires a few global counters.                                      *
 *                                                                      *
 * Arguments:                                                           *
 * None.
 ************************************************************************/
void
display_results (void)
{
    /**********************************************************************/
    double log_time_diff;                       /* Time difference calculations */
    unsigned int nbr_days = 0;                  /* How many days in this run? */

    const char *visits1 = "             Visits Per:           New Visitor:  %8lu  Old Visitor:  %8lu\n";
    const char *pages1 = "              Pages Per:           New Visitor:  %8lu  Old Visitor:  %8lu\n";

    /**********************************************************************/

    /* Calculate the number of days in this run */
    if (current_time == 0 || run_start_time == 0) {
        nbr_days = 1;
    } else {
        log_time_diff = difftime (current_time, run_start_time) + DAY_IN_SECONDS;
        nbr_days = (unsigned int) floor (log_time_diff / DAY_IN_SECONDS);
        if (nbr_days <= 0)
            nbr_days = 1;
    }

    if (flag_raw_stats_only == true) {
        VPRINT (VERBOSE1, "Days\tVisitors\tNewVisitors\tOldVisitors\t   Visits\tNewVisits\tOldVisits\t   Pages\tNewPages\tOldPages%s\n", "");
        printf ("%4u\t%8lu\t%8lu\t%8lu\t%8lu\t%8lu\t%8lu\t%8lu\t%8lu\t%8lu\n", nbr_days, nbr_visitors, nbr_new_visitors, nbr_visitors - nbr_new_visitors, nbr_visits, nbr_visits_new, nbr_visits_old, nbr_pages_viewed, nbr_pages_new, nbr_pages_old);
        if (bad_lines > 0) {
            VPRINT (VERBOSE1, "%s%8lu of %10lu\n", msg_I_number_bad_lines, bad_lines, total_lines);
        }

    } else {

        /* Display the results */
        printf ("Visits/ors:    Visitors: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_visitors, nbr_new_visitors, nbr_visitors - nbr_new_visitors);
        printf ("Per Day (%3d): Visitors: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_days, nbr_visitors / nbr_days, nbr_new_visitors / nbr_days, (nbr_visitors - nbr_new_visitors) / nbr_days);

        printf ("\n");
        printf ("Visits:    Total Visits: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_visits, nbr_visits_new, nbr_visits_old);
        printf (" Per Day:  Total Visits: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_visits / nbr_days, nbr_visits_new / nbr_days, nbr_visits_old / nbr_days);

        if ((nbr_visitors - nbr_new_visitors) > 0) {
            printf (visits1, nbr_visits_new / nbr_new_visitors, nbr_visits_old / (nbr_visitors - nbr_new_visitors));
        } else {
            printf (visits1, nbr_visits_new / nbr_new_visitors, 0);
        }

        printf ("\n");
        printf ("Pages:      Total Pages: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_pages_viewed, nbr_pages_new, nbr_pages_old);
        printf (" Per Day:   Total Pages: %8lu  New Visitors: %8lu  Old Visitors: %8lu\n", nbr_pages_viewed / nbr_days, nbr_pages_new / nbr_days, nbr_pages_old / nbr_days);

        if ((nbr_visitors - nbr_new_visitors) > 0) {
            printf (pages1, nbr_pages_new / nbr_new_visitors, nbr_pages_old / (nbr_visitors - nbr_new_visitors));
        } else {
            printf (pages1, nbr_pages_new / nbr_new_visitors, 0);
        }
        printf ("\n");
        printf ("Visitors Identified by Cookies: %7lu   -> %.2f%%\n", successful_cookies, (double) ((double) successful_cookies / (double) nbr_visitors) * 100.0);
        printf ("Removed Cookies:                %7lu\n", cleaned_cookies);
        if (bad_lines > 0) {
            VPRINT (VERBOSE0, "%s%8lu of %10lu\n", msg_I_number_bad_lines, bad_lines, total_lines);
        }
    }
}


/************************************************************************
 * cleanup_excess_cookies                                               *
 *                                                                      *
 * We have a linked list of newly added cookies, if appropriate,        *
 * delete cookie entries that are not                                   *
 * marked as good, and haven't been touched in over a Visit Timeout     *
 * period (default of 30 minutes)                                       *
 *                                                                      *
 * Returns the number of cookies removed                                *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 ************************************************************************/
unsigned int
cleanup_excess_cookies (DB ** db_ptr)
{
    /**********************************************************************/
    DBC *dbcurs_ptr = NULL;                     /* DB Pointer for cursor'ing */
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    int db_rtn = 0;                             /* return value from DBD Get */
    visit_data_type *visit_data = NULL;

    unsigned int excess_cookies = 0;            /* How many cookies did we clean/remove? */

    /**********************************************************************/

    VPRINT (VERBOSE3, "CLEANUP COOKIES.%s\n", "");

    /* Acquire a cursor for the database. */
    if ((db_rtn = (*db_ptr)->cursor (*db_ptr, NULL, &dbcurs_ptr, 0)) != 0) {
        (*db_ptr)->err (*db_ptr, db_rtn, "DB->cursor");
        ERRVPRINT (VERBOSE0, msg_F_db_cursor_create, "");
        exit (V_EXIT_DB_CURSOR_CREATION);
    }
    VPRINT (VERBOSE5, "dbcurs_ptr == %p", dbcurs_ptr);

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    while ((db_rtn = dbcurs_ptr->c_get (dbcurs_ptr, &dbt_key, &dbt_data, DB_NEXT)) == 0) {
        visit_data = (visit_data_type *) dbt_data.data;
        if ((visit_data->is_cookie == true) && (visit_data->good_cookie == false) && (difftime (current_time, visit_data->last_access) > g_visittimeout)) {
            VPRINT (VERBOSE5, "  DeleteMe: %s\n", (char *) dbt_key.data);
            del_hash (db_ptr, (char *) dbt_key.data);
            excess_cookies++;
        }
    }
    db_rtn = dbcurs_ptr->c_close (dbcurs_ptr);

    return (excess_cookies);
}


/************************************************************************
 * cookie_exists                                                        *
 *                                                                      *
 * Have we seen this cookie?                                            *
 * If yes, return 1 and fill in remaining data fields                   *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 * cookie_type * cookie         Expects the "cookie" entry to be        *
 *                                filled. This is the cookie to then    *
 *                                look for.                             *
 ************************************************************************/
inline int
cookie_exists (DB ** db_ptr, cookie_type * cookie)
{
    /**********************************************************************/
    int rtn = 0;                                /* Return value */
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    int db_rtn;                                 /* return value from DBD Get */

    /**********************************************************************/

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    dbt_key.data = (char *) cookie->cookie_key;
    dbt_key.size = (strlen (cookie->cookie_key) + 1) * sizeof (char);

    db_rtn = (*db_ptr)->get (*db_ptr, NULL, &dbt_key, &dbt_data, 0);
    if (db_rtn != 0) {
        rtn = 0;
    } else {
        VPRINT (VERBOSE5, "  Cookie Exists: Key retrieved: %s\n", (char *) dbt_key.data);
        memcpy (&cookie->data, dbt_data.data, dbt_data.size);
        rtn = 1;
    }
    return (rtn);
}


/************************************************************************
 * address_browser_exists                                               *
 *                                                                      *
 * Have we seen this address_browser?                                   *
 * If yes, return 1 and fill in remaining data fields                   *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 * address_browser_type * addrbrwsr     Expects the                     *
 *                                "address_browser_key" entry to be     *
 *                                filled. This is the key to then       *
 *                                look for.                             *
 ************************************************************************/
inline int
address_browser_exists (DB ** db_ptr, address_browser_type * addrbrwsr)
{
    /**********************************************************************/
    int rtn = 0;                                /* Return value */
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    int db_rtn;                                 /* return value from DBD Get */

    /**********************************************************************/

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    dbt_key.data = (char *) addrbrwsr->address_browser_key;
    dbt_key.size = (strlen (addrbrwsr->address_browser_key) + 1) * sizeof (char);;

    db_rtn = (*db_ptr)->get (*db_ptr, NULL, &dbt_key, &dbt_data, 0);
    if (db_rtn != 0) {
        rtn = 0;
    } else {
        VPRINT (VERBOSE5, "  Address Exists: Key retrieved: %s\n", (char *) dbt_key.data);
        memcpy (&addrbrwsr->data, dbt_data.data, dbt_data.size);
        rtn = 1;
    }
    return (rtn);
}


/************************************************************************
 * store_cookie                                                         *
 *                                                                      *
 * Store this cookie into the hash DB.                                  *
 * Compact any strings down to minimum levels.                          *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 * cookie_type * cookie         This is the cookie and associated       *
 *                                values to store.                      *
 ************************************************************************/
inline void
store_cookie (DB ** db_ptr, cookie_type * cookie)
{
    /**********************************************************************/
    void *data_ptr;                             /* pointer to the data store to store in the hash */
    int data_size;                              /* size of fixed data in the cookie */
    int addbr_size;                             /* size of variable sized data in the cookie */
    int key_size;                               /* size of the cookie itself - key for the hash */

    /**********************************************************************/

    VPRINT (VERBOSE3, "  STORE COOKIE: %s\n", cookie->cookie_key);
    data_size = sizeof (visit_data_type);
    addbr_size = ((strlen (cookie->address_browser) + 1) * sizeof (char));

    /* Grab the memory needed for all the cookie's data */
    data_ptr = xmalloc (data_size + addbr_size);

    memcpy (data_ptr, &cookie->data, sizeof (visit_data_type));
    memcpy (data_ptr + data_size, &cookie->address_browser, addbr_size);
    key_size = (strlen (cookie->cookie_key) + 1) * sizeof (char);

    add_hash (db_ptr, &cookie->cookie_key, key_size, data_ptr, data_size + addbr_size);
    free (data_ptr);                            /* release the memory grabbed */
}


/************************************************************************
 * store_address_browser                                                *
 *                                                                      *
 * Store this address browser into the hash DB.                         *
 * Compact any strings down to minimum levels.                          *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 * address_browser_type * addrbrwsr     This is the address/brswr       *
 *                                Combo and associated values to store. *
 ************************************************************************/
inline void
store_address_browser (DB ** db_ptr, address_browser_type * addrbrwsr)
{
    /**********************************************************************/
    void *data_ptr;                             /* pointer to the data store to store in the hash */
    int key_size;                               /* size of the cookie itself - key for the hash */

    /**********************************************************************/

    VPRINT (VERBOSE3, "  STORE ADDR/BRWSR: %s\n", addrbrwsr->address_browser_key);
    VPRINT (VERBOSE5, "    Start: %ld\tLast Access: %ld\tPages: %d\tVisits: %d\n", addrbrwsr->data.visit_start, addrbrwsr->data.last_access, addrbrwsr->data.nbr_pages, addrbrwsr->data.nbr_visits);

    /* Grab the memory needed for all the address/browser data */
    data_ptr = XMALLOC (visit_data_type, 1);

    memcpy (data_ptr, &addrbrwsr->data, sizeof (visit_data_type));
    key_size = (strlen (addrbrwsr->address_browser_key) + 1) * sizeof (char);

    add_hash (db_ptr, &addrbrwsr->address_browser_key, key_size, data_ptr, sizeof (visit_data_type));
    free (data_ptr);                            /* release the memory grabbed */
}


/************************************************************************
 * cleanup_old_records                                                  *
 *                                                                      *
 * Removes all records older than the given number of days.             *
 * Will move the current time to the end of the current day for         *
 *   ensuring that a given days records are removed, and not part       *
 *   thereof.                                                           *
 *                                                                      *
 * Assumes that g_cleanup_mode has been set to a value > 0              *
 *                                                                      *
 * Arguments:                                                           *
 * DB ** db_ptr                 Pointer to the Database to operate on.  *
 ************************************************************************/
void
cleanup_old_records (DB ** db_ptr)
{
    /**********************************************************************/
    DBC *dbcurs_ptr = NULL;                     /* DB Pointer for cursor'ing */
    DBT dbt_key, dbt_data;                      /* Key and Data Pointers for results from BDB */
    int db_rtn = 0;                             /* return value from DBD Get */
    visit_data_type *visit_data = NULL;

    time_t days_end;

    long unsigned int counter = 0;

    /**********************************************************************/

    VPRINT (VERBOSE1, "%s %d %s\n", "Do Cleanup! Records Older Than:", g_cleanup_mode, "Days");

    /* Acquire a cursor for the database. */
    if ((db_rtn = (*db_ptr)->cursor (*db_ptr, NULL, &dbcurs_ptr, 0)) != 0) {
        (*db_ptr)->err (*db_ptr, db_rtn, "DB->cursor");
        ERRVPRINT (VERBOSE0, msg_F_db_cursor_create, "");
        exit (V_EXIT_DB_CURSOR_CREATION);
    }
    VPRINT (VERBOSE5, "dbcurs_ptr == %p", dbcurs_ptr);

    memset (&dbt_key, 0, sizeof (dbt_key));
    memset (&dbt_data, 0, sizeof (dbt_data));

    days_end = time_end_today();
    while ((db_rtn = dbcurs_ptr->c_get (dbcurs_ptr, &dbt_key, &dbt_data, DB_NEXT)) == 0) {
        visit_data = (visit_data_type *) dbt_data.data;
        if (difftime (days_end, visit_data->last_access) / DAY_IN_SECONDS >= g_cleanup_mode) {
            VPRINT (VERBOSE5, "  DeleteMe: %s\n", (char *) dbt_key.data);
            del_hash (db_ptr, (char *) dbt_key.data);
            counter++;
        }
    }

    VPRINT (VERBOSE1, "Removed %lu records\n", counter);

    db_rtn = dbcurs_ptr->c_close (dbcurs_ptr);
}


/************************************************************************
 * time_end_today                                                       *
 *                                                                      *
 * Return the time in time_t format for the end of the current day.     *
 *                                                                      *
 * Arguments:                                                           *
 * NONE                                                                 *
 ************************************************************************/
time_t
time_end_today (void)
{
    /**********************************************************************/
    time_t now;
    struct tm *days_end_tm;
    time_t days_end;
    /**********************************************************************/

    /* Set the time to compare to be the end of the current day.
     *  This helps get a complete days worth of records */
    now = time (NULL);
    days_end_tm = localtime (&now);
    days_end_tm->tm_sec = 59;
    days_end_tm->tm_min = 59;
    days_end_tm->tm_hour = 23;
    days_end = mktime (days_end_tm);
    VPRINT (VERBOSE2, "The end of today is: %s --> %ld\n", asctime (days_end_tm), days_end);

    return (days_end);

}


/************************************************************************
 ************************************************************************
 *                      END OF FILE                                     *
 ************************************************************************
 ************************************************************************/
